from typing import Union
from typing_extensions import TypeAlias

from openai.types.chat import (
    ChatCompletionToolMessageParam,
    ChatCompletionUserMessageParam,
    ChatCompletionSystemMessageParam,
    ChatCompletionFunctionMessageParam,
)

from nexusflowai.types.chat_completion_assistant_message_param import (
    NexusflowAIChatCompletionAssistantMessageParam,
)

NexusflowAIChatCompletionMessageParam: TypeAlias = Union[
    ChatCompletionSystemMessageParam,
    ChatCompletionUserMessageParam,
    NexusflowAIChatCompletionAssistantMessageParam,
    ChatCompletionToolMessageParam,
    ChatCompletionFunctionMessageParam,
]



from typing import Any, Dict, Optional, Tuple, Union

import json

from pydantic import BaseModel

from datamodel_code_generator.format import PythonVersion
from datamodel_code_generator.parser.jsonschema import JsonSchemaParser
from datamodel_code_generator.model import get_data_model_types

from datamodel_code_generator import DataModelType


def json_schema_to_dataclasses(
    json_schema: str, with_import: bool = False
) -> Optional[str]:
    target_python_version = PythonVersion.PY_38
    output_model_type = DataModelType.DataclassesDataclass
    data_model_types = get_data_model_types(output_model_type, target_python_version)

    parser = JsonSchemaParser(
        source=json_schema,
        data_model_type=data_model_types.data_model,
        data_model_root_type=data_model_types.root_model,
        data_model_field_type=data_model_types.field_model,
        data_type_manager_type=data_model_types.data_type_manager,
        use_schema_description=True,
        use_field_description=True,
    )

    try:
        results = parser.parse(with_import=with_import)
    except:  # pylint: disable=bare-except # pragma: no cover
        return None

    if not results:  # pragma: no cover
        return None
    elif isinstance(results, str):
        bodies = [results]
    else:
        bodies = [result.body for _, result in sorted(results.items())]

    parser_results = parser.results

    res = "\n\n".join(bodies)
    return res, parser_results


def try_convert_to_dataclasses_str(
    input_: Union[BaseModel, Dict[str, Any]], with_import: bool = False
) -> Tuple[Optional[str], Optional[str]]:
    """
    Returns
    -------
    json_schema_str: str
    dataclasses_str: str

    """
    if isinstance(input_, type) and issubclass(input_, BaseModel):
        input_: Dict[str, Any] = input_.model_json_schema()

    if not isinstance(input_, Dict):
        return None, None, None

    if input_.get("type") == "json_schema":
        json_schema = input_["json_schema"]
        # OpenAI strict style schema
        if "name" in json_schema:
            outer_name = input_["json_schema"]["name"]
            input_ = input_["json_schema"]["schema"]
            input_["title"] = outer_name
        else:  # NexusflowAI mimic format
            input_ = input_["json_schema"]

    is_object_type = input_.get("type") == "object"
    has_allOf = input_.get("allOf") is not None
    has_ref = input_.get("$ref") is not None
    assert (
        is_object_type or has_allOf or has_ref
    ), f"Input JSON schema top-level type must be an object, but got `{input_.get('type')}`"

    try:
        str_input_ = json.dumps(input_)
    except json.JSONDecodeError:  # pragma: no cover
        return None, None, None

    res, datamodel_fields = json_schema_to_dataclasses(
        json_schema=str_input_, with_import=with_import
    )
    if not res:  # pragma: no cover
        return None, None, None

    if isinstance(input_, Dict):
        # For some reason, the datamodel_code_generator logic turns "default": None into str 'None'
        res = res.replace("'None'", "None")

    return str_input_, res, datamodel_fields



from nexusflowai.types.completion_usage import NexusflowAICompletionUsage
from nexusflowai.types.completion_choice import NexusflowAICompletionChoice
from nexusflowai.types.completion import NexusflowAICompletion
from nexusflowai.types.chat_completion_message_tool_call import (
    NexusflowAIChatCompletionMessageToolCall,
)
from nexusflowai.types.chat_completion_message import NexusflowAIChatCompletionMessage
from nexusflowai.types.chat_completion import NexusflowAIChatCompletion
from nexusflowai.types.chat_completion_content_part_image_bytes_param import (
    NexusflowAIChatCompletionContentPartImageBytesParam,
)
from nexusflowai.types.chat_completion_chunk import NexusflowAIChatCompletionChunk
from nexusflowai.types.chat_completion_create_params import (
    NexusflowAICompletionCreateParams as NexusflowAIChatCompletionCreateParams,
)



from dataclasses import dataclass

from openai.types.completion_create_params import CompletionCreateParams
from nexusflowai.validators.base_input_validator import BaseInputValidator


@dataclass
class CompletionsInputValidator(BaseInputValidator):
    def validate(self) -> CompletionCreateParams:
        self._validate_stream(params=self.request_body)

        self._validate_stop_token(params=self.request_body)

        self._check_for_unsupported_parameter_value(
            params=self.request_body,
            parameter_name="best_of",
            exception_class=NotImplementedError,
            supported_value=1,
            assign_supported_value=True,
        )
        self._check_for_unsupported_parameter_value(
            params=self.request_body,
            parameter_name="echo",
            exception_class=NotImplementedError,
            supported_value=False,
            assign_supported_value=True,
        )
        self._check_for_unsupported_parameter_value(
            params=self.request_body,
            parameter_name="n",
            exception_class=NotImplementedError,
            supported_value=1,
            assign_supported_value=True,
        )

        self._check_for_unsupported_parameter_value(
            params=self.request_body,
            parameter_name="logprobs",
            exception_class=NotImplementedError,
            exception_message="Logprobs is not supported at this time",
        )

        self._validate_temperature(params=self.request_body)

        return self.request_body



from __future__ import annotations

import os
import asyncio
from typing import Union, Mapping
from typing_extensions import override

import httpx
from openai import resources, _exceptions
from openai._qs import Querystring
from openai._types import (
    NOT_GIVEN,
    Omit,
    Timeout,
    NotGiven,
    Transport,
    ProxiesTypes,
    RequestOptions,
)
from openai._utils import is_given, is_mapping
from openai._version import __version__
from openai._streaming import Stream
from openai._streaming import AsyncStream
from openai._exceptions import APIStatusError
from openai._base_client import DEFAULT_MAX_RETRIES, SyncAPIClient, AsyncAPIClient

from nexusflowai.resources import (
    Completions,
    AsyncCompletions,
    CompletionsWithRawResponse,
    AsyncCompletionsWithRawResponse,
    Chat,
    AsyncChat,
    ChatWithRawResponse,
    AsyncChatWithRawResponse,
    Models,
    AsyncModels,
    ModelsWithRawResponse,
    AsyncModelsWithRawResponse,
)

from nexusflowai._exceptions import NexusflowAIError


__all__ = [
    "Timeout",
    "Transport",
    "ProxiesTypes",
    "RequestOptions",
    "resources",
    "NexusflowAI",
    "AsyncNexusflowAI",
    "Client",
    "AsyncClient",
]


class NexusflowAI(SyncAPIClient):
    completions: Completions
    chat: Chat
    models: Models
    with_raw_response: NexusflowAIWithRawResponse

    # client options
    api_key: str
    organization: str | None

    def __init__(
        self,
        *,
        api_key: str | None = None,
        organization: str | None = None,
        base_url: str | httpx.URL | None = None,
        timeout: Union[float, Timeout, None, NotGiven] = NOT_GIVEN,
        max_retries: int = DEFAULT_MAX_RETRIES,
        default_headers: Mapping[str, str] | None = None,
        default_query: Mapping[str, object] | None = None,
        # Configure a custom httpx client. See the [httpx documentation](https://www.python-httpx.org/api/#client) for more details.
        http_client: httpx.Client | None = None,
        # Enable or disable schema validation for data returned by the API.
        # When enabled an error APIResponseValidationError is raised
        # if the API responds with invalid data for the expected schema.
        #
        # This parameter may be removed or changed in the future.
        # If you rely on this feature, please open a GitHub issue
        # outlining your use-case to help us decide if it should be
        # part of our public interface in the future.
        _strict_response_validation: bool = False,
    ) -> None:
        """Construct a new synchronous openai client instance.

        This automatically infers the following arguments from their corresponding environment variables if they are not provided:
        - `api_key` from `NEXUSFLOWAI_API_KEY`
        - `organization` from `NEXUSFLOWAI_ORG_ID` - Defaults to `None`
        """
        if api_key is None:
            api_key = os.environ.get("NEXUSFLOWAI_API_KEY")
        if api_key is None:
            raise NexusflowAIError(
                "The api_key client option must be set either by passing api_key to the client or by setting the NEXUSFLOWAI_API_KEY environment variable"
            )
        self.api_key = api_key

        if organization is None:
            organization = os.environ.get("NEXUSFLOWAI_ORG_ID")
        self.organization = organization

        if base_url is None:
            base_url = os.environ.get("NEXUSFLOWAI_BASE_URL")
        if base_url is None:
            base_url = "https://api.nexusflow.ai/v1"

        super().__init__(
            version=__version__,
            base_url=base_url,
            max_retries=max_retries,
            timeout=timeout,
            http_client=http_client,
            custom_headers=default_headers,
            custom_query=default_query,
            _strict_response_validation=_strict_response_validation,
        )

        self._default_stream_cls = Stream

        self.completions = Completions(self)
        self.chat = Chat(self)
        self.models = Models(self)
        self.with_raw_response = NexusflowAIWithRawResponse(self)

    @property
    @override
    def qs(self) -> Querystring:  # pragma: no cover
        return Querystring(array_format="comma")

    @property
    @override
    def auth_headers(self) -> dict[str, str]:  # pragma: no cover
        api_key = self.api_key
        return {"Authorization": f"Bearer {api_key}"}

    @property
    @override
    def default_headers(self) -> dict[str, str | Omit]:  # pragma: no cover
        return {
            **super().default_headers,
            "NexusflowAI-Organization": (
                self.organization if self.organization is not None else Omit()
            ),
            **self._custom_headers,
        }

    def copy(
        self,
        *,
        api_key: str | None = None,
        organization: str | None = None,
        base_url: str | httpx.URL | None = None,
        timeout: float | Timeout | None | NotGiven = NOT_GIVEN,
        http_client: httpx.Client | None = None,
        max_retries: int | NotGiven = NOT_GIVEN,
        default_headers: Mapping[str, str] | None = None,
        set_default_headers: Mapping[str, str] | None = None,
        default_query: Mapping[str, object] | None = None,
        set_default_query: Mapping[str, object] | None = None,
    ) -> NexusflowAI:  # pragma: no cover
        """
        Create a new client instance re-using the same options given to the current client with optional overriding.

        It should be noted that this does not share the underlying httpx client class which may lead
        to performance issues.
        """
        if default_headers is not None and set_default_headers is not None:
            raise ValueError(
                "The `default_headers` and `set_default_headers` arguments are mutually exclusive"
            )

        if default_query is not None and set_default_query is not None:
            raise ValueError(
                "The `default_query` and `set_default_query` arguments are mutually exclusive"
            )

        headers = self._custom_headers
        if default_headers is not None:
            headers = {**headers, **default_headers}
        elif set_default_headers is not None:
            headers = set_default_headers

        params = self._custom_query
        if default_query is not None:
            params = {**params, **default_query}
        elif set_default_query is not None:
            params = set_default_query

        http_client = http_client or self._client
        return self.__class__(
            api_key=api_key or self.api_key,
            organization=organization or self.organization,
            base_url=base_url or str(self.base_url),
            timeout=self.timeout if isinstance(timeout, NotGiven) else timeout,
            http_client=http_client,
            max_retries=max_retries if is_given(max_retries) else self.max_retries,
            default_headers=headers,
            default_query=params,
        )

    # Alias for `copy` for nicer inline usage, e.g.
    # client.with_options(timeout=10).foo.create(...)
    with_options = copy

    def __del__(self) -> None:  # pragma: no cover
        if not hasattr(self, "_has_custom_http_client") or not hasattr(self, "close"):
            # this can happen if the '__init__' method raised an error
            return

        if self._has_custom_http_client:
            return

        self.close()

    @override
    def _make_status_error(
        self,
        err_msg: str,
        *,
        body: object,
        response: httpx.Response,
    ) -> APIStatusError:  # pragma: no cover
        data = body.get("error", body) if is_mapping(body) else body
        if response.status_code == 400:
            return _exceptions.BadRequestError(err_msg, response=response, body=data)

        if response.status_code == 401:
            return _exceptions.AuthenticationError(
                err_msg, response=response, body=data
            )

        if response.status_code == 403:
            return _exceptions.PermissionDeniedError(
                err_msg, response=response, body=data
            )

        if response.status_code == 404:
            return _exceptions.NotFoundError(err_msg, response=response, body=data)

        if response.status_code == 409:
            return _exceptions.ConflictError(err_msg, response=response, body=data)

        if response.status_code == 422:
            return _exceptions.UnprocessableEntityError(
                err_msg, response=response, body=data
            )

        if response.status_code == 429:
            return _exceptions.RateLimitError(err_msg, response=response, body=data)

        if response.status_code >= 500:
            return _exceptions.InternalServerError(
                err_msg, response=response, body=data
            )
        return APIStatusError(err_msg, response=response, body=data)


class AsyncNexusflowAI(AsyncAPIClient):
    completions: AsyncCompletions
    chat: AsyncChat
    models: AsyncModels
    with_raw_response: AsyncNexusflowAIWithRawResponse
    # client options
    api_key: str
    organization: str | None

    def __init__(
        self,
        *,
        api_key: str | None = None,
        organization: str | None = None,
        base_url: str | httpx.URL | None = None,
        timeout: Union[float, Timeout, None, NotGiven] = NOT_GIVEN,
        max_retries: int = DEFAULT_MAX_RETRIES,
        default_headers: Mapping[str, str] | None = None,
        default_query: Mapping[str, object] | None = None,
        # Configure a custom httpx client. See the [httpx documentation](https://www.python-httpx.org/api/#asyncclient) for more details.
        http_client: httpx.AsyncClient | None = None,
        # Enable or disable schema validation for data returned by the API.
        # When enabled an error APIResponseValidationError is raised
        # if the API responds with invalid data for the expected schema.
        #
        # This parameter may be removed or changed in the future.
        # If you rely on this feature, please open a GitHub issue
        # outlining your use-case to help us decide if it should be
        # part of our public interface in the future.
        _strict_response_validation: bool = False,
    ) -> None:
        """Construct a new async openai client instance.

        This automatically infers the following arguments from their corresponding environment variables if they are not provided:
        - `api_key` from `NEXUSFLOWAI_API_KEY`
        - `organization` from `NEXUSFLOWAI_ORG_ID` - Defaults to `None`
        """
        if api_key is None:
            api_key = os.environ.get("NEXUSFLOWAI_API_KEY")
        if api_key is None:
            raise NexusflowAIError(
                "The api_key client option must be set either by passing api_key to the client or by setting the NEXUSFLOWAI_API_KEY environment variable"
            )
        self.api_key = api_key

        if organization is None:
            organization = os.environ.get("NEXUSFLOWAI_ORG_ID")
        self.organization = organization

        if base_url is None:
            base_url = os.environ.get("NEXUSFLOWAI_BASE_URL")
        if base_url is None:
            base_url = "https://api.nexusflow.ai/v1"

        super().__init__(
            version=__version__,
            base_url=base_url,
            max_retries=max_retries,
            timeout=timeout,
            http_client=http_client,
            custom_headers=default_headers,
            custom_query=default_query,
            _strict_response_validation=_strict_response_validation,
        )

        self._default_stream_cls = AsyncStream

        self.completions = AsyncCompletions(self)
        self.chat = AsyncChat(self)
        self.models = AsyncModels(self)
        self.with_raw_response = AsyncNexusflowAIWithRawResponse(self)

    @property
    @override
    def qs(self) -> Querystring:  # pragma: no cover
        return Querystring(array_format="comma")

    @property
    @override
    def auth_headers(self) -> dict[str, str]:  # pragma: no cover
        api_key = self.api_key
        return {"Authorization": f"Bearer {api_key}"}

    @property
    @override
    def default_headers(self) -> dict[str, str | Omit]:  # pragma: no cover
        return {
            **super().default_headers,
            "NexusflowAI-Organization": (
                self.organization if self.organization is not None else Omit()
            ),
            **self._custom_headers,
        }

    def copy(
        self,
        *,
        api_key: str | None = None,
        organization: str | None = None,
        base_url: str | httpx.URL | None = None,
        timeout: float | Timeout | None | NotGiven = NOT_GIVEN,
        http_client: httpx.AsyncClient | None = None,
        max_retries: int | NotGiven = NOT_GIVEN,
        default_headers: Mapping[str, str] | None = None,
        set_default_headers: Mapping[str, str] | None = None,
        default_query: Mapping[str, object] | None = None,
        set_default_query: Mapping[str, object] | None = None,
    ) -> AsyncNexusflowAI:  # pragma: no cover
        """
        Create a new client instance re-using the same options given to the current client with optional overriding.

        It should be noted that this does not share the underlying httpx client class which may lead
        to performance issues.
        """
        if default_headers is not None and set_default_headers is not None:
            raise ValueError(
                "The `default_headers` and `set_default_headers` arguments are mutually exclusive"
            )

        if default_query is not None and set_default_query is not None:
            raise ValueError(
                "The `default_query` and `set_default_query` arguments are mutually exclusive"
            )

        headers = self._custom_headers
        if default_headers is not None:
            headers = {**headers, **default_headers}
        elif set_default_headers is not None:
            headers = set_default_headers

        params = self._custom_query
        if default_query is not None:
            params = {**params, **default_query}
        elif set_default_query is not None:
            params = set_default_query

        http_client = http_client or self._client
        return self.__class__(
            api_key=api_key or self.api_key,
            organization=organization or self.organization,
            base_url=base_url or str(self.base_url),
            timeout=self.timeout if isinstance(timeout, NotGiven) else timeout,
            http_client=http_client,
            max_retries=max_retries if is_given(max_retries) else self.max_retries,
            default_headers=headers,
            default_query=params,
        )

    # Alias for `copy` for nicer inline usage, e.g.
    # client.with_options(timeout=10).foo.create(...)
    with_options = copy

    def __del__(self) -> None:  # pragma: no cover
        if not hasattr(self, "_has_custom_http_client") or not hasattr(self, "close"):
            # this can happen if the '__init__' method raised an error
            return

        if self._has_custom_http_client:
            return

        try:
            asyncio.get_running_loop().create_task(self.close())
        except Exception:  # pylint: disable=broad-exception-caught
            pass

    @override
    def _make_status_error(
        self,
        err_msg: str,
        *,
        body: object,
        response: httpx.Response,
    ) -> APIStatusError:  # pragma: no cover
        data = body.get("error", body) if is_mapping(body) else body
        if response.status_code == 400:
            return _exceptions.BadRequestError(err_msg, response=response, body=data)

        if response.status_code == 401:
            return _exceptions.AuthenticationError(
                err_msg, response=response, body=data
            )

        if response.status_code == 403:
            return _exceptions.PermissionDeniedError(
                err_msg, response=response, body=data
            )

        if response.status_code == 404:
            return _exceptions.NotFoundError(err_msg, response=response, body=data)

        if response.status_code == 409:
            return _exceptions.ConflictError(err_msg, response=response, body=data)

        if response.status_code == 422:
            return _exceptions.UnprocessableEntityError(
                err_msg, response=response, body=data
            )

        if response.status_code == 429:
            return _exceptions.RateLimitError(err_msg, response=response, body=data)

        if response.status_code >= 500:
            return _exceptions.InternalServerError(
                err_msg, response=response, body=data
            )
        return APIStatusError(err_msg, response=response, body=data)


class NexusflowAIWithRawResponse:
    def __init__(self, client: NexusflowAI) -> None:
        self.completions = CompletionsWithRawResponse(client.completions)
        self.chat = ChatWithRawResponse(client.chat)
        self.models = ModelsWithRawResponse(client.models)


class AsyncNexusflowAIWithRawResponse:
    def __init__(self, client: AsyncNexusflowAI) -> None:
        self.completions = AsyncCompletionsWithRawResponse(client.completions)
        self.chat = AsyncChatWithRawResponse(client.chat)
        self.models = AsyncModelsWithRawResponse(client.models)


Client = NexusflowAI

AsyncClient = AsyncNexusflowAI



from typing import Dict, Any, Union, Literal

from openai._types import Headers

ALLOWED_MODELS = Literal["nexus-tool-use-20240816"]
CLIENT_HEADER_KEY = "X-Client-ID"
CLIENT_NAME = "NexusflowAI"


def get_extra_header() -> Dict[str, Any]:
    return {CLIENT_HEADER_KEY: CLIENT_NAME}


def transform_client_header(extra_headers: Union[Headers, None]) -> Headers:
    if not extra_headers:
        return get_extra_header()

    return extra_headers | get_extra_header()



from logging import getLogger

# We import logger here first since submodules will reference it
logger = getLogger("NexusflowAI")

# pylint: disable=wrong-import-position
from nexusflowai.client import NexusflowAI, AsyncNexusflowAI



from typing import Any, Callable, Dict, List, Optional, Tuple, Union

from enum import Enum

from dataclasses import asdict, dataclass

from itertools import chain, repeat

import json

import ast

from uuid import uuid4

from openai.types.chat import ChatCompletionToolParam
from openai.types.chat.completion_create_params import ResponseFormat
from openai.types.chat.chat_completion_tool_choice_option_param import (
    ChatCompletionNamedToolChoiceParam,
)

from nexusflowai.types import NexusflowAIChatCompletionMessageToolCall
from nexusflowai.types.chat_completion_message_tool_call import (
    Function as ChatCompletionFunctionCall,
)

from nexusflowai.validators.create_with_tools.utils import (
    CodeExecutionHelper,
    CleanedFunction,
    get_tool_choice_prefix,
    resolve_tools_and_tool_choice,
    clean_raw_response,
)
from nexusflowai.validators.create_with_tools.preprocess import (
    RavenFunctionCall,
    FunctionCallTranslator,
    RavenQuery,
    ToolsDefs,
)
from nexusflowai.validators.json_schema_to_dataclasses import (
    try_convert_to_dataclasses_str,
)


def _collect_factory(
    fun_name: str,
    fun_calls: List[RavenFunctionCall],
    fun_args: Union[Dict[str, Any], None],
    fun_par_idx: int,
    par_fun_calls: List[List[RavenFunctionCall]],
) -> Callable:
    def _collect(*args, **kwargs) -> RavenFunctionCall:
        if fun_args is not None:
            fun_args_dict = fun_args.copy()
            new_kwargs = dict()

            for arg_key, arg in chain(zip(repeat(None), args), kwargs.items()):
                # First case: we encountered either a positional arg (i.e. arg_key is None), or a hallucinated keyword arg.
                if arg_key not in fun_args_dict:
                    # If this is an "extraneous" argument, i.e. all known fun params have already been matched to args, then just silently drop the arg.
                    if len(fun_args_dict) <= 0:
                        continue

                    # Clobber non-hallucinated arg keyword names that occur after the hallucinated arg.
                    new_arg_key = next(iter(fun_args_dict))

                # Second case: this is a keyword that matches one of the known params given in the fun def.
                else:
                    new_arg_key = arg_key

                new_kwargs[new_arg_key] = arg
                fun_args_dict.pop(new_arg_key)

            # Any remaining missing required params should be added as keyword args set to their "default" values.
            new_kwargs |= {
                arg_name: arg_dict["def"]
                for arg_name, arg_dict in fun_args_dict.items()
                if arg_dict["req"]
            }

            kwargs = new_kwargs

        elif args:
            raise TypeError("Positional args are not allowed in Raven function calls")

        fun_call = RavenFunctionCall(fun_name, kwargs)

        fun_calls.append(fun_call)

        while fun_par_idx >= len(par_fun_calls):
            par_fun_calls.append([])
        par_fun_calls[fun_par_idx].append(fun_call)

        return fun_call

    return _collect


@dataclass
class RavenFunctionNameVisitor(ast.NodeVisitor):
    def __init__(self) -> None:
        super().__init__()
        self.fun_names = []
        self.fun_par_idxs = []
        self.par_idx = 0

    def visit_Module(self, node: ast.Module) -> None:
        for sub_node in node.body:
            self.visit(sub_node)
            self.par_idx += 1

    def visit_Call(self, node: ast.Call) -> None:
        if isinstance(node.func, ast.Name):
            self.fun_names.append(node.func.id)
            self.fun_par_idxs.append(self.par_idx)
        self.generic_visit(node)


class FunctionCallResponseTranslator:
    def get_function_names(self, tree: ast.AST) -> Tuple[List[str], List[int]]:
        fun_name_visitor = RavenFunctionNameVisitor()
        fun_name_visitor.visit(tree)
        return fun_name_visitor.fun_names, fun_name_visitor.fun_par_idxs

    def get_tool_defs(self, tools: List[ChatCompletionToolParam]) -> List[str]:
        tool_defs = []
        for tool in tools:
            for tool in tools:
                tool_params = tool.get("function", {}).get("parameters", {})
                if "$defs" in tool_params:
                    tool_defs.extend(list(tool_params["$defs"].keys()))
        return tool_defs

    def raw_response_to_tool_calls(
        self,
        tools: List[ChatCompletionToolParam],
        raw_response: str,
        tool_choice: ChatCompletionNamedToolChoiceParam | None = None,
    ) -> Tuple[List[NexusflowAIChatCompletionMessageToolCall], str, str]:
        tools, processed_tool_choice = resolve_tools_and_tool_choice(tools, tool_choice)
        cfs = list(map(CleanedFunction.from_tool_param, tools))

        dirty_function_name_to_cf = {cf.dirty_name: cf for cf in cfs}
        tool_choice_prefix = get_tool_choice_prefix(
            processed_tool_choice, dirty_function_name_to_cf
        )
        if tool_choice_prefix is not None:
            raw_response = f"{tool_choice_prefix}{raw_response}"

        clean_function_name_to_cf = {cf.clean_name: cf for cf in cfs}
        tools = [cf.clean_tool_param(tool) for cf, tool in zip(cfs, tools)]

        fun_name_to_args = self.maybe_parse_fun_args(tools)
        rft = ResponseFormatTranslator()
        tool_defs_ctx = rft.get_tool_defs_context(tools)

        response = clean_raw_response(raw_response)
        fun_calls, par_fun_calls = self.parse_function_calls(
            response, fun_name_to_args, tool_defs_ctx
        )
        content = None
        if not fun_calls:
            return [], response, content

        tool_calls = []
        for fun_calls in par_fun_calls:  # pylint: disable=redefined-argument-from-local
            if len(fun_calls) <= 0:
                continue

            fun_call = fun_calls[0]

            cf = clean_function_name_to_cf.get(fun_call.name)
            name, kwargs = fun_call.name, fun_call.kwargs
            if cf is not None:
                name, kwargs = cf.dirty_function_call_dict(name, kwargs)

            chat_fc = {"name": name}
            try:
                ResponseFormatTranslator._resolve_enums(
                    kwargs
                )  # pylint: disable=protected-access
                chat_fc["arguments"] = json.dumps(kwargs)
            except Exception as e:
                raise e

            tool_call = NexusflowAIChatCompletionMessageToolCall(
                id=f"call_{str(uuid4()).replace('-', '')}",
                type="function",
                function=ChatCompletionFunctionCall.model_validate(chat_fc),
                execution_result=None,
            )
            tool_calls.append(tool_call)

        if tool_choice in ["required_with_chat", "none"]:
            chat_content_strs = []
            updated_tool_calls = []
            for tool_call in tool_calls:
                if tool_call.function.name == "chat":
                    chat_content_strs.append(
                        json.loads(tool_call.function.arguments)["message"]
                    )
                else:
                    updated_tool_calls.append(tool_call)

            tool_calls = updated_tool_calls or None
            content = "\n".join(chat_content_strs) or None

        return tool_calls, response, content

    def maybe_parse_fun_args(
        self, tools: List[ChatCompletionToolParam]
    ) -> Dict[str, Any]:
        fun_name_to_args = dict()

        for tool in tools:
            fun = tool["function"]
            fun_name = fun["name"]

            req_arg_names = set(fun["parameters"].get("required", []))

            args = {
                arg_name: {
                    "req": arg_name in req_arg_names,
                    "def": arg_dict.get("default", None),
                }
                for arg_name, arg_dict in fun["parameters"]
                .get("properties", dict())
                .items()
            }

            fun_name_to_args[fun_name] = args

        return fun_name_to_args

    def parse_function_calls(
        self,
        source: str,
        fun_name_to_args: Dict[str, Dict[str, Any]],
        tool_defs_ctx: Dict[str, Any],
    ) -> Tuple[List[RavenFunctionCall], List[List[RavenFunctionCall]]]:
        fun_calls = []
        par_fun_calls = []

        ceh = CodeExecutionHelper()
        root_source, tree = ceh.clean_input(source)
        if tree is None:
            return fun_calls, par_fun_calls

        ceh.check_security(tree)

        fun_names, fun_par_idxs = self.get_function_names(tree)

        root_tree = ast.parse(root_source)
        for par_idx, tree in enumerate(root_tree.body):
            if isinstance(tree.value, ast.Call) and tree.value.func.id in tool_defs_ctx:
                source = ast.unparse(tree)
                # pylint: disable=eval-used
                defs_obj = eval(source, tool_defs_ctx)
                defs_obj_dump = asdict(defs_obj)

                fn_call = RavenFunctionCall(
                    name=tree.value.func.id, kwargs=defs_obj_dump
                )

                fun_calls.append(fn_call)
                par_fun_calls.append([fn_call])

            else:
                env = dict()
                for fun_name, fun_par_idx in zip(fun_names, fun_par_idxs):
                    if par_idx != fun_par_idx:
                        continue

                    fun_args = fun_name_to_args.get(fun_name, None)
                    if fun_args is not None:
                        env[fun_name] = _collect_factory(
                            fun_name, fun_calls, fun_args, fun_par_idx, par_fun_calls
                        )
                    else:
                        env[fun_name] = lambda *args, **kwargs: None

                source = ast.unparse(tree)

                try:
                    exec(source, env)  # pylint: disable=exec-used
                except:  # pylint: disable=bare-except
                    pass

        return fun_calls, par_fun_calls


class ResponseFormatTranslator:
    def get_tool_defs_context(
        self,
        tools: List[ChatCompletionToolParam],
    ) -> Dict[str, Any]:
        tool_defs = [
            tool.tool
            for tool in FunctionCallTranslator().format_tools(tools)
            if isinstance(tool, ToolsDefs)
        ]

        if not tool_defs:
            return {}

        ceh = CodeExecutionHelper()

        tool_def_names = [tool["function"]["name"] for tool in tool_defs]
        for tool in tool_defs:
            tool_defs_list = tool["function"]["parameters"]["$defs"]
            tool_def_names.extend(list(tool_defs_list.keys()))

        tool_defs_strs = RavenQuery.construct_tool_defs_strs(
            tool_defs, with_import=True
        )

        context = dict()

        for tool_defs_str in tool_defs_strs:
            _, tree = ceh.clean_input(tool_defs_str)
            if tree is None:
                return None

            ceh.check_security(tree)
            try:
                # pylint: disable=exec-used
                exec(tool_defs_str, context)
            except:
                # pylint: disable=raise-missing-from
                raise RuntimeError("Failed to parse dataclasses_str")

        context = {k: v for k, v in context.items() if k in tool_def_names}

        return context

    def raw_response_to_parsed(
        self,
        response_format: ResponseFormat,
        raw_response: str,
    ) -> Optional[Dict[str, Any]]:
        raw_response = clean_raw_response(raw_response)

        ceh = CodeExecutionHelper()
        raw_response, tree = ceh.clean_input(raw_response)
        if tree is None:
            return None

        ceh.check_security(tree)

        _, dataclasses_str, parser_results = try_convert_to_dataclasses_str(
            response_format, with_import=True
        )

        _, tree = ceh.clean_input(dataclasses_str)
        ceh.check_security(tree)

        model_response = raw_response.removeprefix("extract_item(value=").removesuffix(
            ")"
        )

        context = dict()
        try:
            # pylint: disable=exec-used
            exec(dataclasses_str, context)
        except:
            # pylint: disable=raise-missing-from
            raise RuntimeError("Failed to parse dataclasses_str")

        try:
            # pylint: disable=eval-used
            python_res = eval(model_response, context)
        except:
            # pylint: disable=raise-missing-from
            raise RuntimeError("Failed to parse model response")

        json_res = asdict(python_res)

        self._resolve_enums(json_res)

        aliases = {}
        for parser_result in parser_results:
            for field in parser_result.fields:
                if field.alias:
                    aliases[field.name] = field.alias
        self._replace_field_names(json_res, aliases)

        return json_res

    @classmethod
    def _resolve_enums(cls, obj: Any) -> None:
        """
        Destructively replaces content in object that are enums into their values.
        """
        if isinstance(obj, dict):
            for k in list(obj):
                if isinstance(obj[k], Enum):
                    obj[k] = obj[k].value
                else:
                    cls._resolve_enums(obj[k])
        elif isinstance(obj, list):
            for i, obj_i in enumerate(obj):
                if isinstance(obj_i, Enum):
                    obj[i] = obj_i.value
                else:
                    cls._resolve_enums(obj_i)

    def _replace_field_names(self, obj: Any, aliases: Dict[str, str]) -> None:
        if not aliases:
            return

        if isinstance(obj, dict):
            for k in list(obj):
                self._replace_field_names(obj[k], aliases)
                if k in aliases:
                    obj[aliases[k]] = obj.pop(k)
        elif isinstance(obj, list):
            for obj_i in obj:
                self._replace_field_names(obj_i, aliases)



from typing import Union, Optional, Iterable
from typing_extensions import Literal, Required

from openai.types.chat.completion_create_params import CompletionCreateParamsBase

from nexusflowai.types.chat_completion_message_param import (
    NexusflowAIChatCompletionMessageParam,
)


class NexusflowAICompletionCreateParamsBase(CompletionCreateParamsBase):
    messages: Required[Iterable[NexusflowAIChatCompletionMessageParam]]


class NexusflowAICompletionCreateParamsNonStreaming(
    CompletionCreateParamsBase, total=False
):
    stream: Optional[Literal[False]]


class NexusflowAICompletionCreateParamsStreaming(CompletionCreateParamsBase):
    stream: Required[Literal[True]]


NexusflowAICompletionCreateParams = Union[
    NexusflowAICompletionCreateParamsNonStreaming,
    NexusflowAICompletionCreateParamsStreaming,
]
